#!/usr/bin/env ./localpython/bin/python

import os
import sys
import numpy
import tensor
import sptensor
import tenmat
import scipy
import scipy.stats
import scipy.special
import scipy.optimize
import pyroc
import time
import scipy
import scipy.io
import tensor_fac
import commands


MY_HDFS_DIR = '.'

''' kernel parameter '''
class KernelParam:
    def __init__(self,name):
        self.name = name
        self.kreg = 1e-6
        self.tol = 1e-3
        self.nu = 0.5 #matern cov
        self.phi = 1.0 #matern cov
        self.alpha = 1.0 #matern cov
        self.gamma = 1 #rbf cov
        self.degree = 2 #poly cov
        self.taper = 'spherical' #taper function
        self.theta = 1.0 #taper function


''' object to store gram matrix and their eigen values/vectors (bigger ones)'''
class Kernel:
    def __init__(self):
        self.eig_vals = None
        self.eig_vecs = None
        self.K = None

''' Object to store kernel function for tensor'''
class TensorKernel:
    def __init__(self, nmod):
        self.nmod = nmod
        self.K = [None for i in range(nmod)]


class Model:
    def __init__(self):
        self.var_total = 50
        self.stream_total = 5000
        '''learning rate related '''
        self.kai = 1
        self.tao0 = 4096
        self.eta0 = 0.1
        self.subsample_siz = None
        self.dim = 3
        self.offset = 2.0
        ''' learning strategy
            0: update all modes;
            1: update each mode cyclicly
            2: update all modes, with L1
            3: update each mode cyclicly, with L1
        '''
        self.strategy = 1
        self.lam = 0.1

class Data:
    def __init__(self):
        self.nmod = None
        self.nvec = None
        self.Y = None
        self.nzind = None
        self.test_ind = None
        self.test_y = None


''' Caclulate the gram matrix of the n*p input matrix, where each row is a data point'''
def kernel_func(X, param):
    n,p = X.shape
    K = None
    if param.name == 'linear':
        K = numpy.dot(X,X.transpose())
    elif param.name == 'rbf':
        norm2 = numpy.sum(X**2,axis=1)
        norm2 = numpy.tile(norm2, [n,1])
        K = norm2 + norm2.transpose() - 2*numpy.dot(X, X.transpose())
        K = numpy.exp(-param.gamma*K)
    elif param.name == 'poly':
        K = numpy.dot(X,X.transpose())**param.degree
    elif param.name == 'taper_matern':
        if param.taper == 'spherical' and param.nu>0.5:
            print 'For spherical taper function, nu must <=0.5'
            return None
        if param.taper == 'wendland1' and param.nu>1.5:
            print 'For Wendland1 taper function, nu must <=1.5'
            return None
        if param.taper == 'wendland2' and param.nu>2.5:
            print 'For Wendland2 taper function, nu must <=2.5'
            return None
        norm2 = numpy.sum(X**2,axis=1)
        norm2 = numpy.tile(norm2, [n,1])
        norm2 = norm2 + norm2.transpose() - 2*numpy.dot(X, X.transpose())
        norm2[norm2<0] = 0.0
        H = numpy.sqrt(norm2).real.reshape(norm2.size)

        if param.taper == 'spherical':
            covTap = ((H < param.theta)*(1 - H/param.theta))**2*(1 + H/(2*param.theta))
        elif param.taper == 'wendland1':
            covTap = ((H < param.theta)*(1 - H/param.theta))**4*(1 + 4*H/param.theta)
        elif param.taper == 'wendland2':
            covTap = ((H < param.theta)*(1 - H/param.theta))**6*(1 + 6*H/param.theta + 35*H**2/(3*param.theta**2))
        else:
            print 'Unkown tapering function'
            return None
        nzH = H[covTap>1e-8] + 1e-8
        nzCovMatern = param.phi/(2**(param.nu-1)*scipy.special.gamma(param.nu))*(param.alpha*nzH)**param.nu*scipy.special.kv(param.nu, param.alpha*nzH)
        nzCovMatern[numpy.isnan(nzCovMatern)] = 1.0
        K = numpy.zeros(n*n)
        K[covTap>1e-8] = nzCovMatern*covTap[covTap>1e-8]
        K = K.reshape([n,n])
    else:
        print 'Unkown kernel types'
        return None

    K = K + param.kreg*numpy.eye(n)

    #here we compress the eigen-decomposition of K by throwing out the smallest eigen values and eigen vectors
    [eig_vals, eig_vecs] = numpy.linalg.eig(K)
    eig_val_threshold  = numpy.max(eig_vals)*param.tol
    ret = Kernel()
    ret.K = K
    ret.eig_vals = eig_vals[eig_vals>eig_val_threshold]
    ret.eig_vals = ret.eig_vals.reshape([ret.eig_vals.size, 1])
    ret.eig_vecs = eig_vecs[:,eig_vals>eig_val_threshold]
    return ret

'''X must be a list of arrays, each arry is a N_k*p_k latent representation of the nodes in k-th mode '''
def tensor_kernel_func(X, param):
    ret = TensorKernel(len(X))
    for i in range(ret.nmod):
        ret.K[i] = kernel_func(X[i],param)
    return ret

def taper_matern_grad_U(U, param):
    n,dim = U.shape
    norm2 = numpy.sum(U**2,axis=1)
    norm2 = numpy.tile(norm2, [n,1])
    norm2 = norm2 + norm2.transpose() - 2*numpy.dot(U, U.transpose())
    norm2[norm2<0] = 0.0
    H = numpy.sqrt(norm2).real.reshape(norm2.size)

    tapered = (H < param.theta)*(1 - H/param.theta)
    if param.taper == 'spherical':
        covTap = tapered**2*(1 + H/(2*param.theta))
        covTapGrad = -2.0/param.theta*tapered*(1+H/(2*param.theta)) + tapered**2/(2*param.theta)
    elif param.taper == 'wendland1':
        covTap = tapered**4*(1.0 + 4*H/param.theta)
        covTapGrad = -4.0/param.theta*tapered**3*(1+4*H/param.theta) + tapered**4*4/param.theta
    elif param.taper == 'wendland2':
        covTap = tapered**6*(1 + 6*H/param.theta + 35*H**2/(3*param.theta**2))
        covTapGrad = -6.0/param.theta*tapered**5*(1.0+6*H/param.theta+35*H**2/3/param.theta**2) + tapered**6*(6.0/param.theta+70*H/3/param.theta**2)
    else:
        print 'Unkown tapering function'
        return None

    #nzH = H[covTap>1e-8] + 1e-8
    nzH = H[H<param.theta] + 1e-8
    nzH = nzH*(numpy.abs(nzH)>1e-8)
    nzCovMatern = param.phi/(2**(param.nu-1)*scipy.special.gamma(param.nu))*(param.alpha*nzH)**param.nu*scipy.special.kv(param.nu, param.alpha*nzH)
    nzCovMatern[numpy.isnan(nzCovMatern)] = 1.0
    nzCovMatern[numpy.isinf(nzCovMatern)] = 1.0


    bkgrad = -0.5*(scipy.special.kv(param.nu-1,param.alpha*nzH) + scipy.special.kv(param.nu+1,param.alpha*nzH))
    bkgrad[numpy.isnan(bkgrad)] = 1.0
    bkgrad[numpy.isinf(bkgrad)] = 1.0

    nzCovMaternGrad = param.phi/(2**(param.nu-1)*scipy.special.gamma(param.nu)) \
            * (param.nu*param.alpha*(param.alpha*nzH)**(param.nu-1)*scipy.special.kv(param.nu,param.alpha*nzH) \
            + param.alpha*(param.alpha*nzH)**param.nu*bkgrad)
    nzCovMaternGrad[numpy.isnan(nzCovMaternGrad)] = 0.0
    nzCovMaternGrad[numpy.isinf(nzCovMaternGrad)] = 0.0

    gKh = numpy.zeros(n*n)
    gKh[H<param.theta] = nzCovMaternGrad*covTap[H<param.theta] + nzCovMatern*covTapGrad[H<param.theta]
    gKh = gKh.reshape([n,n])

    gdU = [None for i in range(dim)]
    for i in range(dim):
        repU = numpy.tile(U[:,i].reshape(n,1),[1,n])
        coefs =  (repU - repU.transpose()).reshape(n*n)
        #coefs[H>1e-8] = coefs[H>1e-8]/H[H>1e-8]
        #coefs[H<=1e-8] = 0.0
        coefs = coefs/(H+1e-8)
        coefs = coefs.reshape([n,n])
        gdU[i] = coefs*gKh
    return gdU

''' Ukvec: a long vector (array) represent \U_k
    k: k-th mode
    Utmsnok: a tensor
    eigval_list: a list, where each element is a array of eigen values of gram-matrix of each mode
    eigvec_k: eigen vectors of k-th mode
    pram: kernel parameters
'''
def tensor_gradient_U(Ukvec, k, Utmsnok,  eigval_list, eigvec_k, param):
    nvec = Utmsnok.shape
    n = Utmsnok.size()
    nmod = len(nvec)
    nk = nvec[k]

    U_k = Ukvec.reshape([nk, Ukvec.size/nk])
    kernel_k = kernel_func(U_k, param)
    inv_k_half = numpy.dot(numpy.dot(kernel_k.eig_vecs, numpy.diagflat(kernel_k.eig_vals**(-0.5))), kernel_k.eig_vecs.transpose())
    sqroot_term = Utmsnok.ttm(inv_k_half,k)

    core = tensor.tensor([1],[1 for i in range(nmod)])
    DD = core.ttm(eigval_list, [i for i in range(nmod)])
    DD = tensor.tensor(DD.data/(1.0+DD.data), DD.shape)
    inv_eigvals = [1.0/x.transpose() for x in eigval_list]
    KV = numpy.linalg.solve(kernel_k.K, eigvec_k)
    inv_eigvals[k] = numpy.dot(eigvec_k.transpose(),KV).diagonal().reshape(eigval_list[k].shape[::-1])

    fun_val = 0.5*n/nk*numpy.sum(numpy.log(kernel_k.eig_vals))  + 0.5*numpy.sum(sqroot_term.data**2) + \
            0.5*numpy.sum(DD.ttm(inv_eigvals,[i for i in range(nmod)]).data)

    g = numpy.zeros(U_k.shape)
    d = U_k.shape[1]
    A = tenmat.tenmat(Utmsnok,[k]).data
    KM = numpy.linalg.solve(kernel_k.K, numpy.dot(A,A.transpose()))
    modnok = numpy.array(range(nmod))
    modnok = modnok[modnok!=k]
    if param.name == 'taper_matern':
        gradU = taper_matern_grad_U(U_k, param)
    for j in range(0,d):
        repU = numpy.tile(U_k[:,j].reshape([nk,1]),[1,nk])
        if param.name == 'rbf':
            Kd = -2*param.gamma*numpy.linalg.solve(kernel_k.K, kernel_k.K*(repU.transpose() - repU))
        elif param.name == 'linear':
            Kd = numpy.linalg.solve(kernel_k.K, repU)
        elif param.name == 'poly':
            Kd = param.degree*numpy.linalg(kernel_k.K, kernel_k.K/numpy.dot(U_k, U_k.transpose())*repU)
        elif param.name == 'taper_matern':
            Kd = numpy.linalg.solve(kernel_k.K, gradU[j])
        else:
            print 'Unkown kernel types'
            return None
        b = DD.ttm(inv_eigvals, modnok).data.reshape([eigval_list[k].size,1])
        gCol = n/nk*Kd.diagonal().reshape(nk,1) - numpy.sum(KM.transpose()*Kd, axis=0).reshape(nk,1) - \
               numpy.dot(numpy.dot(eigvec_k.transpose(),Kd).transpose() * KV, b)
        g[:,j] = gCol.reshape(nk).real
    g = g + param.kreg
    g = g/numpy.sqrt(numpy.sum(g**2))
    return [fun_val.real,g.real]



''' sample a small tensor, with siz,
    return a list, each entry is a array containing the selected index
'''
def subsample(data, nvec, siz, strategy):
    subinds = [None for i in range(len(nvec))]

    ''' Just do uniformly sampling,later different strategies will be incorporated '''
    log_prob = 0.0
    for k in range(len(nvec)):
        perm = numpy.random.permutation(nvec[k])
        subinds[k] = perm[0:siz[k]]
        log_prob = log_prob - numpy.sum(numpy.log(range(nvec[k]-siz[k]+1 , nvec[k]+1)))

    return [subinds, log_prob]



''' E step for infiniteTucker
    M: n-d array, the latent value for each tensor entry
    Y: n-d array, the observed tensor entry value
    model: model setting atributes
    model.offset: the offset for probit model
    tensor_kern: the calculated TensorKernel object
'''
def var_E(M, Y, tensor_kern, model):
    tol = 1e-10
    nvec = Y.shape
    nmod = len(nvec)
    nmod_arr = [i for i in range(nmod)]
    nmod_one = [1 for i in range(nmod)]
    eigval_list = [x.eig_vals for x in tensor_kern.K]
    eigvec_list = [x.eig_vecs for x in tensor_kern.K]
    eigvecT_list = [x.eig_vecs.transpose() for x in tensor_kern.K]
    core = tensor.tensor([1],nmod_one)
    DD = core.ttm(eigval_list, nmod_arr)
    DD = tensor.tensor(DD.data/(1.0+DD.data), DD.shape)
    if model.loss == 'Gauss':
        M = tensor.tensor(Y.ttm(eigvecT_list, nmod_arr).tondarray()*DD.data).ttm(eigvec_list, nmod_arr).data.real
    else:
        for iter_num in range(model.var_total):
            P = (2*Y - 1)*scipy.stats.norm.pdf(M-model.offset)/(scipy.stats.norm.cdf((2*Y-1)*(M-model.offset)) + tol)
            Z = tensor.tensor(P + M, P.shape)
            A = Z.ttm(eigvecT_list, nmod_arr).data*DD.data
            A = tensor.tensor(A,A.shape)
            M = A.ttm(eigvec_list, nmod_arr).data.real
            #print iter_num

    return M




def batch_test(data, model, test_sub, subsize, n, ker_param):
    aver_vals = numpy.zeros(len(test_sub))
    U = data.U
    for i in xrange(n):
        [subinds, dummy] = subsample(data, data.nvec, subsize, 0)
        subinds = [numpy.union1d(subinds[k], test_sub[:,k]) for k in xrange(data.nmod)]
        subsample_siz = [len(subinds[k]) for k in xrange(len(subinds))]
        U_sub = [U[i][subinds[i],:]   for i in xrange(data.nmod)]
        nz_ind =  data.nzind
        nz_vals = data.nzvals
        for k in range(data.nmod):
            bool_indicator = numpy.in1d(nz_ind[:,k], subinds[k])
            nz_ind = nz_ind[bool_indicator,:]
            nz_vals = nz_vals[bool_indicator]
            if nz_ind.size == 0:
                #print 'empty sub-tensor!'
                break
        #get the reverse map
        rmap = [None for i in range(data.nmod)]
        for k in range(data.nmod):
            rmap[k] = numpy.zeros(numpy.max(subinds[k]) + 1)
            rmap[k][subinds[k]] = numpy.arange(subinds[k].size)
        ind_sub = numpy.zeros(nz_ind.shape)
        for k in range(data.nmod):
            ind_sub[:,k] = rmap[k][nz_ind[:,k]]
        Y_sub_tensor = sptensor.sptensor(ind_sub, nz_vals.reshape([ind_sub.shape[0],1]), subsample_siz)
        Y_sub = Y_sub_tensor.totensor().data
        if nz_ind.size == 0:
            M_sub = numpy.zeros(subsample_siz)
        else:
            if model.loss == 'Gauss':
                M_sub = Y_sub
            else:
                tfac = tensor_fac.TensorFac()
                [core, U_sub_tmp] = tfac.tucker_als(Y_sub_tensor , model.dim)
                M_sub = core.ttm(U_sub_tmp, numpy.arange(len(U_sub_tmp))).data
                #M_sub = 2*Y_sub - 1

                if nz_ind.size > 0:
                    M_sub = M_sub/numpy.sqrt(numpy.sum(M_sub**2))*numpy.sqrt(numpy.sum(Y_sub))
        #E step
        tensor_kern = tensor_kernel_func(U_sub, ker_param)
        if model.loss == 'Gauss':
            M_sub = var_E(M_sub, Y_sub_tensor.totensor(), tensor_kern, model)
        else:
            M_sub = var_E(M_sub, Y_sub, tensor_kern, model)

        test_sub_ind = numpy.zeros(test_sub.shape)
        for k in xrange(data.nmod):
            test_sub_ind[:,k] = rmap[k][test_sub[:,k]]
        vals = numpy.array([M_sub[tuple(test_sub_ind[i,:])] for i in xrange(test_sub_ind.shape[0])])
        aver_vals += vals

    aver_vals /= n
    return aver_vals



def reduce_job(loss):
    result = {}
    #f = open('tmp.txt','r')
    #for line in f:
    for line in sys.stdin:
        id,y,pred= line.strip().split(' ')
        if id not in result:
            result[id] = [[],[]]
        if loss == 'Gauss':
            result[id][0].append(float(y))
        else:
            result[id][0].append(int(float(y)))
        result[id][1].append(float(pred))

    vals = {}
    for item in result.items():
        id = item[0]
        y = item[1][0]
        pred = item[1][1]
        if loss != 'Gauss':
            rz = [(y[i], pred[i]) for i in range(len(y))]
            roc = pyroc.ROCData(rz)
            vals[ id ] = roc.auc()
            print 'test %s, auc = %g'%(id, vals[ id ] )
        else:
            vals[ id ] = numpy.mean((numpy.array(y)-numpy.array(pred))**2)
            print 'test %s, mse = %g'%(id, vals[ id ])
        
    #get mean & std
    print 'mean = %g , std = %g'%(numpy.mean(vals.values()), numpy.std(vals.values())/numpy.sqrt(len(vals)))



#obj should be numpy ndarray
def put_hdfs_data(hdfs_path, obj):
    os.system('hadoop fs -rm ' + hdfs_path)
    tmp_name = hdfs_path.split('/')[-1]
    os.system('rm -f ' + tmp_name)
    numpy.save(tmp_name, obj)
    os.system('hadoop fs -put ' + tmp_name + ' ' + hdfs_path)
    os.system('rm -f ' + tmp_name)


def load_hdfs_data(hdfs_path):
    tmp_name  = hdfs_path.split('/')[-1]
    os.system('rm -f ' + tmp_name)
    os.system('hadoop fs -get ' + hdfs_path + ' ' + tmp_name)
    ret = numpy.load(tmp_name)
    os.system('rm -f ' + tmp_name)
    return ret


def mapper_job(U_hdfs_path, tr_hdfs_path, loss, subsize, subnum):
    data = Data()
    U = load_hdfs_data(U_hdfs_path)
    #U = [numpy.asarray(x) for x in U]
    tr_data = load_hdfs_data(tr_hdfs_path)
    if  loss == 'Gauss':
        tr_ind = tr_data[0][1:,:].astype(int)
        tr_val = tr_data[1]
    else:
        tr_ind = tr_data[1:,:].astype(int)
        tr_val =  numpy.ones(tr_ind.shape[0])

    nvec = [U[i].shape[0] for i in xrange(U.size)]
    nmod = len(nvec)
    data.nvec = nvec
    data.nmod = nmod
    data.nzind = tr_ind
    data.nzvals = tr_val
    data.U = U

    model = Model()
    model.var_total = 3 
    model.dim = [U[k].shape[1] for k in xrange(data.nmod)]
    model.loss = loss
    #ker_param = KernelParam('taper_matern')
    ker_param = KernelParam('rbf')

    #f = open('test_input','r')
    #for line in f:
    for  line in sys.stdin:
        terms = line.strip().split(' ')
        id = terms[0]
        n = int(terms[1])
        test_subs = numpy.zeros([n, nmod])
        y = numpy.zeros(n)
        st = 2
        for j in xrange(n):
            #take ind, y
            test_subs[j] = numpy.array([int(x) for x in terms[st:st+nmod]])
            if loss == 'Gauss':
                y[j] = float(terms[st+nmod])
            else:
                y[j] = int(float(terms[st+nmod]))
            st = st + nmod + 1
        est_vals = batch_test(data, model, test_subs.astype(int), subsize, subnum, ker_param) 
        for j in xrange(n):
            print '%s %g %g'%(id, y[j], est_vals[j])

    

if __name__ == '__main__':
    numpy.random.seed(0)
    if len(sys.argv) != 7 and len(sys.argv) != 3:
        print 'Usage:'
        print sys.argv[0] + ' -map <U HDFS path> <training file HDFS path> <loss> <subsize> <subnum>'
        print 'e.g., <ibm-u.npy> <ibm-tr.npy> Gauss 50,50,50 10'
        print sys.argv[0] + ' -red <loss>' 
        sys.exit(0)


    if sys.argv[1] == '-map':
        subsize = [int(x) for x in sys.argv[5].split(',')]
        subnum = int(sys.argv[6])
        mapper_job(sys.argv[2], sys.argv[3], sys.argv[4], subsize, subnum)
    elif sys.argv[1] == '-red':
        reduce_job(sys.argv[2])

