# -*- coding: utf-8 -*-

import numpy
import numpy.linalg
import tensor
import tenmat
import sptensor
import sptenmat
import scipy
import scipy.stats
import scipy.optimize
import time


def gradient_Uk(vecU, dim, Xk, Ak, XkAk, AkAk, lam):
    Uk = vecU.reshape(vecU.size/dim,dim)
    f = 0.5*numpy.linalg.norm(Xk - numpy.dot(Uk,Ak))**2
    g = -XkAk + numpy.dot(Uk,AkAk)
    g = g.reshape(g.size)
    return [f,g]

def gradient_core(vecCore, dims, X, U, AX, UU):
    W = tensor.tensor(vecCore.reshape(dims))
    estX = W.ttm(U, numpy.arange(len(U)))
    f = 0.5*numpy.sum((X - estX).data**2)
    WU = W.ttm(UU, numpy.arange(len(U))).data
    g = -AX + WU
    g = g.reshape(g.size)
    return [f,g]


'''
compute argmax || Y X_n A  || where Y is tensor and A is matrix;
the result is simply the the r eigen-vectors with respect to the r larst eigen-values of Y_n*Y_n'
'''
def max_proj(Y, n, r):
    Yn = tenmat.tenmat(Y,[n]).data;
    X = numpy.dot(Yn,Yn.transpose())

    try:
        p = min(max(2*r,20),X.shape[0])
        maxit = max(300, int(numpy.ceil(2.0*X.shape[0]/max(p,1))))
        ''' using this with random init is good for digg1'''
        eigvals,eigvecs = scipy.sparse.linalg.eigsh(X,k=r, ncv=p, maxiter=maxit, which='LM')
        '''this is not as good as eigh or eigs'''
        '''
        eigvals,eigvecs = scipy.sparse.linalg.eigs(X,k=r, ncv=p, maxiter=maxit, which='LM')
        eigvecs = eigvecs.real
        eigvals = eigvals.real
        '''
    except:
        ''' using this with random init is also good '''
        eigvals,eigvecs = scipy.linalg.eigh(X, eigvals=(X.shape[0]-r, X.shape[0]-1))
        pass
    idx = numpy.abs(eigvals).argsort()[::-1]
    A = eigvecs[:,idx[0:r]]

    #flip sign, make the biggest elemnt with postive sign
    max_idx = numpy.argmax(numpy.abs(A), axis=0)
    for i in xrange(r):
        if A[max_idx[i],i]<0:
            A[:,i] = A[:,i]*-1
    return A




class TensorFac:

    def __init__(self, max_iter=10, tol=1e-4, lam=1):
        self.max_iter = max_iter
        self.tol = tol
        self.lam = lam
    '''
    X: sparse tensor
    nvec: latent dimensions
    lam: coefficient for ridge regularization
    return: core: tensor, U, list of ndarray
    '''
    def tucker_als_v1(self, X, nvec):
        #init
        X_dense = X.totensor()
        U = [numpy.random.randn(X.dimsize(n)*nvec[n]).reshape(X.dimsize(n),nvec[n]) for n in xrange(X.ndims())]
        core = tensor.tensor(numpy.random.randn(numpy.array(nvec).prod()).reshape(nvec))
        for iter in xrange(self.max_iter):
            ''' update core '''
            AX = X_dense.ttm([u.transpose() for u in U],numpy.array(range(X.ndims()))).data
            fun = lambda x:gradient_core(x, nvec, X, U, AX, [numpy.dot(u.transpose(),u) for u in U])
            x,f,d = scipy.optimize.fmin_l_bfgs_b(fun, x0=core.data.reshape(core.data.size), args=(), maxfun=100)
            print 'iter=%d, cost =%g'%(iter, f)
            core = tensor.tensor(x.copy().reshape(nvec))
            ''' update each U_i '''
            for k in xrange(X.ndims()):
                #print k
                Xk = sptenmat.sptenmat(X,[k]).tosparsemat()
                modnok = numpy.array(range(X.ndims()))
                modnok = modnok[modnok!=k]
                Ak = core.ttm(U,modnok)
                Ak = tenmat.tenmat(Ak,[k]).data
                AkAk = numpy.dot(Ak,Ak.transpose())
                XkAk = Xk*Ak.transpose() #Xk is sparse matrix
                fun = lambda x:gradient_Uk(x, nvec[k], Xk, Ak, XkAk, AkAk, self.lam)
                x,f,d = scipy.optimize.fmin_l_bfgs_b(fun, x0=U[k].reshape(U[k].size), args=())
                U[k] = x.reshape(U[k].shape)
        return [core, U]


    def tucker_als(self, X, nvec, init='random'):
        U = [None for i in xrange(len(nvec))]
        X = X.totensor()
        Xnorm2 = numpy.sum(X.data**2)
        Xnorm = numpy.sqrt(Xnorm2)

        if init=='random':
            for i in range(len(nvec))[1:]:
                #U[i] = numpy.random.randn(X.dimsize(i),nvec[i])
                U[i] = numpy.random.uniform(size=(X.dimsize(i),nvec[i]))
        else:
            for i in range(len(nvec))[1:]:
                U[i] = max_proj(X, i, nvec[i])
        nmod = numpy.arange(len(nvec))
        fit = 0
        for itr in xrange(self.max_iter):
            fitold = fit
            for k in xrange(len(nvec)):
                Utild = X.ttm(U, nmod[nmod!=k], 't')
                U[k] = max_proj(Utild, k, nvec[k])

            core = Utild.ttm(U,[k], 't')
            normresidual = numpy.sqrt(numpy.abs(Xnorm2 - numpy.sum(core.data**2)))
            fit = 1 - normresidual/Xnorm
            fitchange = numpy.abs(fit - fitold)
            #print 'iter = %d, fit = %g, fitdelta=%g'%(itr, fit, fitchange)

            if itr>1 and fitchange < self.tol:
                break

        return [core,U]
