#!/bin/python

import os
import sys
import commands
from ConfigParser import SafeConfigParser


#directory, file-name, string content
def create_script(folder, name, content):
    with open('%s/%s'%(folder,name),'w') as f:
        f.write(content)

def upload2hdfs(local_file, hdfs_file):
    os.system('hadoop fs -rm -r %s'%hdfs_file)
    os.system('hadoop fs -put %s %s'%(local_file, hdfs_file))

def create_hdfs_dir(name):
    os.system('hadoop fs -rm -r %s'%name)
    os.system('hadoop fs -mkdir %s'%name)

def download_hdfs(source, target):
    os.system('rm -f %s'%target)
    os.system('hadoop fs -get %s %s'%(source, target))


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print 'Usage'
        print sys.argv[0] + ' <input.ini>'
        sys.exit(0)

    conf = SafeConfigParser()
    conf.read(sys.argv[1])

    app_name = conf.get('DinTucker', 'app_name')
    nvec = conf.get('DinTucker', 'size')
    rank = conf.get('DinTucker', 'rank')
    learning_rate  = conf.get('DinTucker', 'learning_rate')
    init_u = conf.get('DinTucker', 'latent_factor_init')
    tensor_tr = conf.get('DinTucker', 'training_tensor')
    subtensor_input = conf.get('DinTucker', 'subtensor_input')
    task = conf.get('DinTucker', 'task')
    if task == 'regression':
        loss = 'Gauss'
    else:
        loss = 'Probit'

    save_intermediate = conf.getint('DinTucker', 'save_intermediate')
    #first, fork a working dir
    app_local = 'app-%s'%app_name
    os.system('rm -rf %s; cp -r app %s'%(app_local,app_local))
    #upload training test, subtensor input, latent factors
    create_hdfs_dir(app_name)
    latent_factor_hdfs = '%s/U.npy'%app_name
    training_hdfs = '%s/train.npy'%app_name
    input_hdfs = '%s/input'%app_name
    upload2hdfs(init_u, latent_factor_hdfs)
    upload2hdfs(tensor_tr, training_hdfs)
    upload2hdfs(subtensor_input, input_hdfs)

    max_iter = conf.getint('DinTucker', 'iter_num')

    mapper_sh = '%s\n./localpython/bin/python -W ignore ./%s/DinTucker.py -map %s %s %s %s %d'%('#!/bin/bash', app_local, training_hdfs, latent_factor_hdfs, loss, learning_rate, max_iter)
    create_script(app_local, 'mapper.sh', mapper_sh)
    red_sh = '%s\n./localpython/bin/python -W ignore ./%s/DinTucker.py -red %s %s %s'%('#!/bin/bash', app_local, nvec, rank, latent_factor_hdfs)
    create_script(app_local, 'reducer.sh', red_sh)
    

    
    hjar = conf.get('HADOOP', 'hadoop_streaming_jar')
    num_mapper = conf.get('HADOOP', 'num_mapper')
    queue = conf.get('HADOOP', 'queue')
    output = '%s/tmp'%app_name
    mapper_timeout = conf.get('HADOOP', 'max_mapper_time')
    python_package = conf.get('HADOOP', 'python_package')
    
    mapper_file = '%s/mapper.sh'%app_local
    red_file = '%s/reducer.sh'%app_local



    cmd = 'hadoop jar %s -D mapred.map.tasks=%s -D mapred.job.queue.name=%s -D mapreduce.task.timeout=%s --files %s -cacheArchive %s#localpython -mapper "%s" -reducer "%s" -numReduceTasks 1 -input %s  -output %s'%(hjar, num_mapper, queue, mapper_timeout, app_local, python_package, mapper_file, red_file, input_hdfs, output)
    
    for i in range(max_iter):
        print 'iter %d'%i
        os.system('hadoop fs -rm -r %s'%output)

        ''' get the global likelihood and calculate gradient '''
        os.system(cmd)
        status, content = commands.getstatusoutput('hadoop fs -cat %s/*'%output)
        print content 

        if save_intermediate == 1:
            download_hdfs(latent_factor_hdfs, '%s/U-iter-%d'%(app_local, i))

    #delete running dir in hadoop
    download_hdfs(latent_factor_hdfs, '%s/result-U.npy'%app_local)
    os.system('hadoop fs -rm -r %s'%app_name)
    print 'Finish!'

