#!/bin/python

import os
import sys
import commands
import numpy as np
from ConfigParser import SafeConfigParser


#directory, file-name, string content
def create_script(folder, name, content):
    with open('%s/%s'%(folder,name),'w') as f:
        f.write(content)

def upload2hdfs(local_file, hdfs_file):
    os.system('hadoop fs -rm -r %s'%hdfs_file)
    os.system('hadoop fs -put %s %s'%(local_file, hdfs_file))

def create_hdfs_dir(name):
    os.system('hadoop fs -rm -r %s'%name)
    os.system('hadoop fs -mkdir %s'%name)

def download_hdfs(source, target):
    os.system('rm -f %s'%target)
    os.system('hadoop fs -get %s %s'%(source, target))




#parse all the test data, generate splits for hadoop streaming
def generate_splits(prefix, num_range, app_dir, entry_num, target_hdfs):
    tmp_input = '%s/tmp.txt'%app_dir 
    with open(tmp_input, 'w') as f:
        for i in num_range:
            test_data = '%s%d.npy'%(prefix, i)
            dt = np.load(test_data)
            ind = dt[0].astype(int)
            y = dt[1]
            st = 0
            for idx in range(entry_num, len(y), entry_num):
                content = '%d %d'%(i, idx - st)
                for j in range(st,idx):
                    content = content + ' ' + ' '.join([str(x) for x in ind[j,:]]) + ' ' + str(y[j])
                f.write(content + '\n')
                st = idx
            if st < len(y):
                content = '%d %d'%(i, len(y) - st)
                for j in range(st, len(y)):
                    content = content + ' ' + ' '.join([str(x) for x in ind[j,:]]) + ' ' + str(y[j])
                f.write(content + '\n')
        
    upload2hdfs(tmp_input, target_hdfs)
    os.system('rm -f %s'%tmp_input)




if __name__ == '__main__':
    if len(sys.argv) != 3:
        print 'Usage'
        print sys.argv[0] + ' <input.ini> <test.ini>'
        sys.exit(0)

    conf = SafeConfigParser()
    test_conf = SafeConfigParser()
    conf.read(sys.argv[1])
    test_conf.read(sys.argv[2])


    app_name = conf.get('DinTucker', 'app_name')
    #nvec = conf.get('DinTucker', 'size')
    #rank = conf.get('DinTucker', 'rank')
    tensor_tr = conf.get('DinTucker', 'training_tensor')
    task = conf.get('DinTucker', 'task')
    if task == 'regression':
        loss = 'Gauss'
    else:
        loss = 'Probit'

    #first, fork a working dir
    app_local = 'app-%s'%app_name
    u_learned = '%s/result-U.npy'%app_local
    #upload training test, subtensor input, latent factors
    create_hdfs_dir(app_name)
    latent_factor_hdfs = '%s/U.npy'%app_name
    training_hdfs = '%s/train.npy'%app_name
    upload2hdfs(u_learned, latent_factor_hdfs)
    upload2hdfs(tensor_tr, training_hdfs)
    #generate input for testing
    bagging_num = test_conf.get('bagging', 'baggin_num')
    subsize = test_conf.get('bagging', 'subtensor_size')
    prefix = test_conf.get('bagging', 'test_file_prefix')
    entry_num = test_conf.getint('bagging', 'entry_num')
    test_file_num = test_conf.getint('bagging', 'test_file_num')
    num_range = range(0, test_file_num)
    input_hdfs = '%s/test_input'%app_name
    generate_splits(prefix, range(0, test_file_num), app_local, entry_num, input_hdfs)


    mapper_sh = '%s\n./localpython/bin/python -W ignore ./%s/evaluate.py -map %s %s %s %s %s'%('#!/bin/bash', app_local, latent_factor_hdfs, training_hdfs,  loss, subsize, bagging_num)
    create_script(app_local, 'mapper-test.sh', mapper_sh)
    red_sh = '%s\n./localpython/bin/python -W ignore ./%s/evaluate.py -red %s'%('#!/bin/bash', app_local, loss)
    create_script(app_local, 'reducer-test.sh', red_sh)
    

    
    hjar = conf.get('HADOOP', 'hadoop_streaming_jar')
    num_mapper = test_conf.get('HADOOP', 'num_mapper')
    queue = conf.get('HADOOP', 'queue')
    output = '%s/tmp'%app_name
    mapper_timeout = conf.get('HADOOP', 'max_mapper_time')
    python_package = conf.get('HADOOP', 'python_package')
    
    mapper_file = '%s/mapper-test.sh'%app_local
    red_file = '%s/reducer-test.sh'%app_local


    cmd = 'hadoop jar %s -D mapred.map.tasks=%s -D mapred.job.queue.name=%s -D mapreduce.task.timeout=%s --files %s -cacheArchive %s#localpython -mapper "%s" -reducer "%s" -numReduceTasks 1 -input %s  -output %s'%(hjar, num_mapper, queue, mapper_timeout, app_local, python_package, mapper_file, red_file, input_hdfs, output)

    os.system('hadoop fs -rm -r %s'%output)
    os.system(cmd)
    status, content = commands.getstatusoutput('hadoop fs -cat %s/*'%output)
    print content 
    
